#!/usr/bin/env python2
### indicates lines added or commented out by Brett Nordgren June, 2014.

print('Importing modules') ###

from setup import *

import numpy as np
import matplotlib.pyplot as plt

from matplotlib.mlab import csd, psd, window_hanning, detrend_mean, detrend_none, cohere
from numpy import pi, square, power, log10, conj


from obspy import read, UTCDateTime
from obspy.signal.invsim import evalresp

import os

def load_mseed( fname ):

    print('loading {0}...'.format(fname))
    st = read(fname)

    gaps = st.getGaps()
    if len(gaps) != 0:
        print('Found gaps: ')
        print(gaps)
        exit()

    st.merge(method=1, interpolation_samples=-1)
    return st[0]


def load_nlnm():

    nlnm = np.loadtxt("models/NLNM.txt", comments = '=', unpack = True)
    freq = 1/nlnm[0]
### amp  = nlnm[3]
    amp = nlnm[1]   ### for short form NLNM.txt 
    return freq, amp

def load_mlnm(): ###

    mlnm = np.loadtxt("models/MLNM.txt", comments = '=', unpack = True) ###
    freq = 1/mlnm[0] ###
    amp = mlnm[1] ### 
    return freq, amp ###

def load_nhnm():
### nhnm = np.loadtxt("models/NHNM.txt", comments = '=', unpack = True, delimiter=', ')
    nhnm = np.loadtxt("models/NHNM.txt", comments = '=', unpack = True) ### for short NHNM.txt:
    freq = 1/nhnm[0]
    amp  = nhnm[1]
    return freq, amp


def resp_type(fname):

    with open(fname, 'r') as f:
        content = f.read()

    if 'ZEROS' in content and 'POLES' in content and 'CONSTANT' in content:
        return 'SAC'

    if 'B050F03' in content:
        return 'RESP'

    return None


def parse_pz(fname):

    with open(fname, 'r') as f:
        content = f.read()

    for line in content.split('\n'):

        if line.startswith('*'):
            continue 

        try:
            tok1, tok2 = line.split()
        except:
            continue

        if tok1 == 'CONSTANT':

            scale = float(tok2)

        elif tok1 == 'ZEROS':

            zeros = [0.0j] * int(tok2)
            ptr = zeros
            pos = 0

        elif tok1 == 'POLES':

            poles = [0.0j] * int(tok2)
            ptr = poles
            pos = 0

        else:

            ptr[pos] = float(tok1) + float(tok2) * 1j
            pos += 1

    return zeros, poles, scale


def calculate_pz(zeros, poles, scale, freq):

    z = 2j * pi * freq
    resp = scale

    for zero in zeros:
        resp *= (z - zero)

    for pole in poles:
        resp /= (z - pole)

    return resp


def abs_squared( z ):
    """
    Returns modulus squared of a complex value.
    """

    assert(np.iscomplexobj(z))

    return square(z.real) + square(z.imag)


def get_cov(data, sample_rate ):
    """
    Returns covariance matrix for signals data
    """

    size = len(data)

    # Create NxN array
    result = []

    for i in range(size):

        result.append([None] * size)

    for i in range(size):

        for j in range(i):

            (Pij, freqs) = csd(data[i], data[j], Fs = sample_rate, **PARAMS)
            result[i][j] = Pij
            result[j][i] = np.conj(Pij)

        (Pii, freqs) = psd(data[i], Fs = sample_rate, **PARAMS)
        result[i][i] = Pii

    result = np.array(result)

    # multiply matrices by omega^4 now, so we don't have to do this later
    result *= power(2 * pi * freqs, 4)

    return (np.array(result), freqs)


def get_cohe(data1, data2, sample_rate):

###    return cohere(data1, data2, Fs = sample_rate, ** PARAMS)
    return cohere(data1, data2, Fs = sample_rate, **PARAMS) ###

def calc_psd( cov_mat, tf ):

    assert(cov_mat.shape[0] == cov_mat.shape[1])

    result = []

    for i in range(cov_mat.shape[0]):

        assert(all(cov_mat[i][i].imag == 0))

        result.append(cov_mat[i][i].real / abs_squared(tf[i]))

    return result


def sleeman( cov_mat, tf ):

    CHANNELS_COUNT = 3

    shape = cov_mat.shape

    if shape[0] != shape[1] or len(shape) != 3:

        raise ValueError("wrong matrix dimensions")

    if shape[0] != CHANNELS_COUNT:

        raise NotImplementedError("no generalization for this case. len(cov_mat) must be equal to 3!")


    def get_noise( ch_id ):

        # Get two other channel id's
        second_id = (ch_id + 1) % 3
        third_id = (ch_id + 2) % 3

        return abs(cov_mat[ch_id][ch_id] - 
                   cov_mat[second_id][ch_id] * cov_mat[ch_id][third_id] /
                   cov_mat[second_id][third_id]) / abs_squared(tf[ch_id])

    return list(map(get_noise, range(CHANNELS_COUNT)))


def holcomb( cov_mat, tf ):

    assert(cov_mat.shape[0] == len(tf) and cov_mat.shape[1] == len(tf))

    CHANNELS_COUNT = len(tf)

    def get_noise( ch_id ):

        other_id = (ch_id + 1) % CHANNELS_COUNT

        return abs(cov_mat[ch_id][ch_id] / (tf[ch_id] * conj(tf[ch_id])) -
                   cov_mat[other_id][ch_id] / (tf[ch_id] * conj(tf[other_id])))

    return list(map(get_noise, range(CHANNELS_COUNT)))



PARAMS = {
    'NFFT'		: 64000,
    'window'	: window_hanning,
    'noverlap'	: 56000,
#    'Fs'		: sample_rate,
    'detrend'	: detrend_none,
    # divide by the sampling frequency so that density function
    # has units of dB/Hz
    'scale_by_freq'	: True,
}


def plot_model(freq, amp, min_freq, max_freq, label):

    mask = (freq >= min_freq)
    x = freq[mask]
    y = amp [mask]

    if XMODE == "seconds":
        x = 1 / x

    plt.plot(x, y, '-', label = label, color = 'black')
   

def plot_densities(freq, densities, min_freq, max_freq, style, label):

    mask = (freq >= min_freq)
    x = freq[mask]

    if XMODE == "seconds":
        x = 1 / x

    for i, amp in enumerate(densities):

        y = 10 * log10(amp[mask])

        plt.plot(
            x, y, style,
            label = label.format(ch_id = i + 1),
            color = COLOR_CYCLE[i],
            linewidth = 1.8 if style is ':' else 0.6
            )

########################################################################

def quake_info(fname):

    path,name=os.path.split(fname)

    mag, dis = name.split('@')
    mag = mag[4:] # delete 'data'

    tmp = []
    for i, c in enumerate(mag):
        if c.isdigit() or c == '.' or c == '+':
            tmp.append(c)
        else:
            break

    inf = ''
    if i < len(mag) - 1:
        inf = mag[i:]

    mag = ''.join(tmp)

    if '3000' in dis:
        dis = 'teleseismic'
    elif '100' in dis: 
        dis = 'regional'
    elif '10' in dis:
        dis = 'local'
    else:
        dis = 'unknown'

    return mag, inf, dis


def quake_data(fname):

    with open(fname, 'r') as f:
        l1 = f.readline().split()
        l2 = f.readline().split()

    assert(len(l1) == len(l2))

    frq = np.array(map(float, l1))
    amp = np.array(map(float, l2))
    amp = 20.0 * np.log10(amp)

    return frq, amp


def plot_quake(frq, amp, text, style, color):

    if XMODE == 'seconds':
        frq = frq.copy()
        frq = 1 / frq

    i = np.argmax(amp)
    x = frq[i]
    y = amp[i]
    plt.text(x, y, text, color=color, verticalalignment='bottom', horizontalalignment='center')
    p = plt.semilogx(frq, amp, style, color=color, lw=2)
    return p

########################################################################

def make_overlap(traces):

    t0 = traces[0]
    t1 = traces[1]
    t2 = traces[2]

    rate0 = t0.stats.sampling_rate
    rate1 = t1.stats.sampling_rate
    rate2 = t2.stats.sampling_rate

    EPS = 1e-5

    if abs(rate0 - rate1) > EPS or abs(rate1 - rate2) > EPS:
        print('Sampling rates differ.')
        exit()

    min0 = t0.stats.starttime
    min1 = t1.stats.starttime
    min2 = t2.stats.starttime

    mint = max(min0, min1)
    mint = max(mint, min2)

    max0 = t0.stats.endtime
    max1 = t1.stats.endtime
    max2 = t2.stats.endtime

    maxt = min(max0, max1)
    maxt = min(maxt, max2)

    t0 = t0.slice(mint, maxt)
    t1 = t1.slice(mint, maxt)
    t2 = t2.slice(mint, maxt)

    if len(t0.data) != len(t1.data) or len(t1.data) != len(t2.data) or len(t0.data) == 0:
        print('Cannot trim traces to the overlapping time.')
        exit()

    traces[0] = t0.data
    traces[1] = t1.data
    traces[2] = t2.data

    return t0.stats


def plot_coherences(freq, cohes, min_freq, max_freq, style, labels):

    mask = (freq >= min_freq)
    x = freq[mask]

    if XMODE == "seconds":
        x = 1 / x

    for i, cohe in enumerate(cohes):

        y = cohe[mask]

        label = labels[i]

        plt.plot(
            x, y, style,
            label = label,
            color = COLOR_CYCLE[i],
            linewidth = 1.8 if style is ':' else 0.6
            )


def main():

    print('Matched file pairs are:')
    for ifile, tfile in zip(DATA_FILES, RESP_FILES):
        print(ifile + ' ' + tfile)

    data = [load_mseed(fname) for fname in DATA_FILES]
    stat = make_overlap(data)

    rate = stat.sampling_rate
    if SAMPLING_RATE > 0:
        rate = SAMPLING_RATE
        print('Using sampling rate from setup.py: {0}'.format(rate))
    else:
        print('Using sampling rate from miniseed: {0}'.format(rate))

    if PLOT_COHE:
        print('Calculating coherence of inputs 1 and 2:')
        cohe12, cfreq = get_cohe(data[0], data[1], rate)
        print('Calculating coherence of inputs 1 and 3:')
        cohe13, cfreq = get_cohe(data[0], data[2], rate)
        print('Calculating coherence of inputs 2 and 3:')
        cohe23, cfreq = get_cohe(data[1], data[2], rate)

        cohes = [ cohe12, cohe13, cohe23 ]


    cov_mat, freqs = get_cov(data, rate)
    # we can remove first element since it will result in NaN anyway
    cov_mat = cov_mat.take(range(1, cov_mat.shape[2]), 2)
    freqs = freqs[1:]


    tf_values = []

    for i, fname in enumerate(RESP_FILES):

        rtype = resp_type(fname)

        if rtype is None:
            raise 'Unknown response type, file ' + fname

        elif rtype == 'SAC':
            zeros, poles, scale = parse_pz(fname)
            resp = calculate_pz(zeros, poles, scale, freqs)

        elif rtype == 'RESP':
            stats = read(DATA_FILES[i])[0].stats
            resp = evalresp(stats.delta, PARAMS['NFFT'], fname, stats.starttime,
                station=stats.station, channel=stats.channel,
                network=stats.network, locid=stats.location, units='DIS')
            resp = resp[1:]

        tf_values.append(resp)


    nhnm_freq, nhnm_amp = load_nhnm()
    nlnm_freq, nlnm_amp = load_nlnm()
    mlnm_freq, mlnm_amp = load_mlnm() ###

    min_freq = min(nlnm_freq)
    max_freq = max(nlnm_freq)

    if CUTOFF != -1:
        if XMODE == 'hertz':
            min_freq = max(CUTOFF, min_freq)
        else:
            min_freq = max(1.0/CUTOFF, min_freq)

    if PLOT_COHE:

        plt.figure()
        plt.xscale('log')
        plot_coherences(cfreq, cohes, min_freq, max_freq, '-', [
            'Coherence between inputs 1 and 2',
            'Coherence between inputs 1 and 3',
            'Coherence between inputs 2 and 3'
        ])

        plt.title('Coherences between channels')

        if XMODE == "hertz":
            xlabel = "log Frequency (Hz)"
        else:
            xlabel = "log Period (seconds)"

        plt.xlabel(xlabel)
        plt.ylabel("Coherence")
        plt.grid()


    # plotting densities

    plt.figure()
    plt.xscale("log")

    if PLOT_NHNM:
        plot_model(nhnm_freq, nhnm_amp, min_freq, max_freq, 'NHNM')

    if PLOT_POWER:
        psd = calc_psd(cov_mat, tf_values)
        plot_densities(freqs, psd, min_freq, max_freq, ':', "Instrument {ch_id} PSD")

    if PLOT_SLEEMAN:
        noise_sleeman = sleeman(cov_mat, tf_values)
        plot_densities(freqs, noise_sleeman, min_freq, max_freq, '-', "Instrument {ch_id} self-noise")

    if PLOT_HOLCOMB:
        noise_holcomb = holcomb(cov_mat, tf_values)
###        plot_densities(freqs, noise_holcomb, min_freq, max_freq, '.', "Instrument {ch_id} self-noise (Holcomb)")
        plot_densities(freqs, noise_holcomb, min_freq, max_freq, '-', "Instrument {ch_id} self-noise (Holcomb)") ###

    if PLOT_MLNM:  ###
        plot_model(mlnm_freq, mlnm_amp, min_freq, max_freq, 'MLNM') ###

    if PLOT_NLNM:
        plot_model(nlnm_freq, nlnm_amp, min_freq, max_freq, 'NLNM')

    if PLOT_LEGEND:
        l1 = plt.legend(loc = 2)
        l1.get_frame().set_alpha(0.1)

##################################################################

    locs = []
    regs = []
    tele = []

    for f in EQFILES:

        mag, inf, dis = quake_info(f)
        frq, amp = quake_data(f)
        mag = 'M' + mag

        if dis == 'local':
            locs.append((frq, amp, mag, inf))
        elif dis == 'regional':
            regs.append((frq, amp, mag, inf))
        elif dis == 'teleseismic':
            tele.append((frq, amp, mag, inf))


    plots = []
    annos = []

    if PLOT_LOCALEQ:
        for frq, amp, mag, inf in locs:
###            p, = plot_quake(frq, amp, mag, '-' , 'black')
            p, = plot_quake(frq, amp, mag, '--' , 'black') ###
        plots.append(p)
        annos.append('Local Earthquakes (~10km)')

    if PLOT_REGIONALEQ:
        for frq, amp, mag, inf in regs:
            p, = plot_quake(frq, amp, mag, '--', 'red')
        plots.append(p)
        annos.append('Regional Earthquakes (~100km)')

    if PLOT_TELESEISMICEQ:
        for frq, amp, mag, inf in tele:
            p, = plot_quake(frq, amp, mag, '--', 'brown') 
        plots.append(p)
  ###   annos.append('Teleseismic Earthquakes (~300km)')
        annos.append('Teleseismic Earthquakes (~3000km)') ###

    if PLOT_LEGEND:
        l2 = plt.legend(plots, annos, loc=4)
        l2.get_frame().set_alpha(0.1)
        plt.gca().add_artist(l1)

##################################################################

    temp = UTCDateTime(stat.endtime - stat.starttime)
    dstr = 'Duration: {0} days, {1}h, {2}m, {3}s'.format(temp.day - 1, temp.hour, temp.minute, temp.second)
    title = TITLE + '\n' + dstr
    plt.title(title)

    if XMODE == "hertz":
###        xlabel = "log Frequency (Hz)"
        xlabel = "Frequency (Hz)" ###       
    else:
###        xlabel = "log Period (seconds)"
        xlabel = "Period (seconds)" ###

    plt.xlabel(xlabel)
### plt.ylabel("Amplitude (dB)")
    plt.ylabel("Noise density - dB vs 1m^2/sec^4 per Hz") ###
    plt.grid()
    plt.show()


if __name__ == '__main__':

    main()

